/*

WinMenu for use in StarDock's WindowBlinds
Copyright (C) 2000 Shane Lardinois (lardino@ibm.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

// The main include code for this DLL
#include "winmenu.h"

// The one and only entry point into this DLL
BOOL APIENTRY DllMain(HANDLE hModule, DWORD  ul_reason_for_call, LPVOID lpReserved)
{
    // Get the reason we are being called
    switch( ul_reason_for_call )
    {
        case DLL_THREAD_ATTACH:
			break;
        case DLL_PROCESS_ATTACH:
			// Save the instance handle
			mHInstance = hModule;

			// Try to locate the bitmap - if the user has specified a bitmap
			// to use in the uic file then use it, otherwise use the default
			// bitmap
			if (!(mMenuButton = CheckProfileString(hModule)))
			{
				// Failed to find a user specified bitmap use the
				// default
				mMenuButton = LoadBitmap(hModule,"#100");
			}
			else
			{
				// We found the bitmap!
			}

			// The the actual information for the bitmap (x & y)
			GetObject(mMenuButton, sizeof(mMenuButtonData),&mMenuButtonData);
			break;
        case DLL_THREAD_DETACH:
			break;
        case DLL_PROCESS_DETACH:
			// Delete the bitmap that we created
			DeleteObject(mMenuButton);
			break;
		default:
			break;
    }

    return TRUE;
} 

// Check the UIS file for a default bitmap to user or if the user
// want to use the default icon for the application
HBITMAP CheckProfileString(HANDLE hInstance)
{
	HMODULE hWBL;
	HBITMAP hReturnBM = NULL;

	// Where our UIS name is going to be copied
	char profilePath[256];
	char bitmapName[256];

	hWBL = GetModuleHandle("WBLIND.DLL");
	
	GetUIS = (GETNAME)GetProcAddress(hWBL,"GetUIS");
	GetWBDIR = (GETNAME)GetProcAddress(hWBL,"GetWBDIR");

	if (GetUIS() == NULL)
	{
		// This is a big time error... couldn't get the filename
		// of the UIS
	}
	else
	{
		// Obtain the UIS filename
		strncpy((LPSTR) profilePath, GetUIS(), 255);

		GetPrivateProfileString("WinMenu","Bitmap","",bitmapName,255,profilePath);

		// If we have a string, see if we can get a bitmap
		if (!strlen(bitmapName))
		{
			// Bitmap name was nothing... nothing else we can do here
		}
		else
		{
			// Get the starting directory along with the bitmap 
			// name to form the entire path
			strncpy(profilePath, GetWBDIR(), 253);
			
			// Make sure we have a backslash someplace
			if ((bitmapName[0] != '\\') && (profilePath[strlen(profilePath) - 1] != '\\'))
			{
				strcat(profilePath,"\\");
			}

			// Cat the rest of the information on the pathname
			strncat(profilePath, bitmapName, 255 - strlen(profilePath));

			// Try to load the bitmap name
			hReturnBM = LoadImage(hInstance, profilePath, IMAGE_BITMAP, 0, 0, LR_LOADFROMFILE);
		}
	}
		
	// Return nothing
	return hReturnBM;
}

DllExport BOOL WBRelease (HWND hWnd, int x, int y, int state)
{
    return TRUE;
}

DllExport BOOL WBPress (HWND hWnd, int x, int y, int state)
{
	// We need to create a menu and associate it with our WND
	HMENU hWinMenu;
	DWORD iMenuSelect;

	// Create the menu item
	hWinMenu = CreatePopupMenu();

	if (!GetInitialConfigKey(hWnd, hWinMenu))
	{
		// Don't show the menu because something underneath us is broken
	}
	else if ((iMenuSelect = TrackPopupMenu ( hWinMenu, TPM_LEFTALIGN | TPM_TOPALIGN | TPM_RETURNCMD | TPM_LEFTBUTTON | TPM_RIGHTBUTTON , x, y, 0 , hWnd, NULL )) != 0)
	{
		// Execute the item that was selected
		ExecuteMenuItem(hWinMenu,iMenuSelect);
	}
	
	// Delete the items and free memory of items in the menu
	DeleteMenuItems(hWinMenu);

	// Destroy the menu since we have already returned!
	DestroyMenu(hWinMenu);

	return TRUE;
}

// Get the initial key for starting a menu... we don't need a submenu item
// in this case
BOOL GetInitialConfigKey(HWND hWnd, HMENU hWinMenu)
{
	BOOL   createKey = TRUE;
	HKEY   currentKey = HKEY_CURRENT_USER; 
	LPCSTR currentString = keyNames[0];
	DWORD  currentCount = 0;
	DWORD  currentCreated; 
	
	// Create or dig until we get to the key we are looking for
	while (createKey)
	{
		if (RegCreateKeyEx( currentKey, currentString, 0, "key", 0, KEY_CREATE_SUB_KEY | KEY_READ, NULL, &currentKey, &currentCreated ) == ERROR_SUCCESS)
		{
			// Move to the next key
			currentCount++;

			// Everything is right in the world
			if (currentCount < (sizeof(keyNames) / sizeof(char *)))
			{
				// Increment to the next key
				currentString = keyNames[currentCount];
			}
			else 
			{
				// We got to the end 
				createKey = FALSE;
			}
		}
		else
		{ 
			// Key doesn't exist or we don't have access - bail out Batman!
			createKey = FALSE;
		}
	}

	// Call to start creating from this key
	createKey = CreateMenuFromRegistry(hWnd, hWinMenu, currentKey);

	// Close this key
	RegCloseKey(currentKey);

	// Return whether we were succesful in getting the key
	return createKey;
}

// Create the menu by getting the our key information from the registery,
// we need the HWND so we can create a bitmap for displaying on the menu
BOOL CreateMenuFromRegistry(HWND hWnd, HMENU hWinMenu, HKEY hKey)
{
	DWORD rKeys;
	DWORD rValues;

	// Recurse looking for subkeys
	rKeys = RecurseForKeys(hWnd, hWinMenu, hKey);
	rValues = RecurseForValues(hWnd, hWinMenu, hKey);

	if ((!rKeys) && (!rValues))
	{
		return FALSE;
	}
	else
	{
		return TRUE;
	}
	
}

// Create subkey items from... sub keys!
BOOL RecurseForKeys(HWND hWnd, HMENU hWinMenu, HKEY hKey)
{
	// Current count of keys
	DWORD currentCount = 0;
	CHAR   keyName[256];
	DWORD  keyNameSize = 256;

	HMENU  hWinSubMenu = NULL;
	HKEY   hSubKey = NULL;

	// Check if we have subkeys, if so create a menu and start at that key
	while (RegEnumKeyEx( hKey, currentCount, keyName, &keyNameSize, 0, NULL, NULL, NULL) == ERROR_SUCCESS)
	{
		if (!(hWinSubMenu = CreateMenu()))
		{
			// Create key didn't succeed
		}
		else if (RegOpenKey(hKey, keyName, &hSubKey) != ERROR_SUCCESS)
		{
			// Open key failed, destroy the menu
			DestroyMenu(hWinSubMenu);
		}
		else if (!CreateMenuFromRegistry(hWnd, hWinSubMenu, hSubKey))
		{
			// Create menu failed or didn't have any items, deleted
			// the menu and the reg key
			DeleteMenuItems(hWinSubMenu);
			DestroyMenu(hWinSubMenu);
		}
		else if (!CreateSubMenuItem(hWnd, hWinMenu, hWinSubMenu, keyName))
		{
			// Create failed, too!?
			DeleteMenuItems(hWinSubMenu);
			DestroyMenu(hWinSubMenu);
		}

		// Close the registry key
		if (hSubKey)
			RegCloseKey(hSubKey);

		// Set/reset our variables
		currentCount++;
		keyNameSize = 256;
	}

	return currentCount;
}
	
// Create values for any sub keys
BOOL RecurseForValues(HWND hWnd, HMENU hWinMenu, HKEY hKey)
{
	// Variables for enumerating the key we found
	CHAR   keyName[256];
	DWORD  keyNameSize = 256;
	CHAR   keyData[256];
	DWORD  keyDataSize = 256;
	DWORD  keyType;
	DWORD  currentCount = 0;

	// See if our key is valid, if so, we can enumerate the values under it to get the
	// items we want to display in the registry

	while (RegEnumValue( hKey, currentCount, keyName, &keyNameSize, NULL, &keyType, NULL, NULL) == ERROR_SUCCESS)
	{
		// If the key is a string, then let's get the data that goes along with it
		if (!(keyType == REG_SZ))
		{
			// The key is the wrong type!?
		}
		else if (RegQueryValueEx( hKey, keyName, NULL, &keyType, (BYTE *)keyData, &keyDataSize ) != ERROR_SUCCESS)
		{
			// Failed to query the value!
		}
		else if (!CreateMenuItem(hWnd, hWinMenu, keyName, keyData))
		{
			// Bad things happened, we failed to add to the
			// menu, lets get out!
			return FALSE;
		}

		// Reset our buffer values
		keyNameSize = 256;
		keyDataSize = 256;

		// Generally a good idea to increment the loop counter
		currentCount++;
	}

	// If we have some items in currentCount then we return TRUE, since the menu
	// is worth showing
	return currentCount;
}

BOOL CreateSubMenuItem(HWND hWnd, HMENU hWinMenu, HMENU hWinSubMenu, LPCSTR keyName)
{
	// The default menu item
	MENUITEMINFO iMenuInfo = { sizeof(MENUITEMINFO), MIIM_STATE | MIIM_TYPE | MIIM_ID | MIIM_SUBMENU, MFT_STRING, MFS_ENABLED , NULL, NULL, NULL, NULL, NULL, NULL, NULL };

	// Set the sub menu
	iMenuInfo.hSubMenu = hWinSubMenu;
	iMenuInfo.wID = ((DWORD)hWinSubMenu & 0x00FF) + (GetMenuItemCount(hWinSubMenu) << 8);

	// Set the name
	iMenuInfo.dwTypeData = (char *)keyName;
	iMenuInfo.cch = strlen(keyName);

	// Insert the sub menu
	return InsertMenuItem(hWinMenu, 0, TRUE, &iMenuInfo);
}

BOOL CreateMenuItem(HWND hWnd, HMENU hWinMenu, LPCSTR keyName, LPCSTR keyValue)
{ 
	// The default menu item
	MENUITEMINFO iMenuInfo = { sizeof(MENUITEMINFO),  MIIM_CHECKMARKS | MIIM_STATE | MIIM_TYPE | MIIM_DATA | MIIM_ID, MFT_STRING, MFS_UNCHECKED | MFS_ENABLED , NULL, NULL, NULL, NULL, NULL, NULL, NULL };

	// Fill in a pointer to the string that is created to hold the action that 
	// the button is going to take is it gets selected, we have to free this
	// before we destroy the menu!
	iMenuInfo.dwItemData = (DWORD)strcpy((char *)malloc((size_t)strlen(keyValue)+1),keyValue);
	iMenuInfo.wID = ((DWORD)hWinMenu & 0x00FF) + (GetMenuItemCount(hWinMenu) << 8);

	// A pointer to the name of the
	iMenuInfo.dwTypeData = (char *)keyName;
	iMenuInfo.cch = strlen(keyName);
		
	// Insert an item into the menu
	return (InsertGraphicalMenuItem( hWnd, hWinMenu, keyValue, &iMenuInfo)); 
}

// The icon from a specific file that was passed in...
BOOL InsertGraphicalMenuItem(HWND hWnd, HMENU hWinMenu, LPCSTR keyValue, MENUITEMINFO *pMenuInfo)
{
	HICON   iIconData;
	
	BOOL    retValue = FALSE;

	char *  parameterIndex;

	// Check if there is a comma in the keyValue, if so we want
	// remove it since it is a parameter
	if ((parameterIndex = strchr(keyValue,',')) != NULL)
	{
		// Insert a string terminator so we only get the filename
		*parameterIndex = '\0';
	}

	// Extract the icon from the file if it exists (first icon)
	iIconData = ExtractIcon(mHInstance, keyValue, 0);

	// Check if we have no image
	if (((DWORD)iIconData != NULL) && ((DWORD)iIconData != 1))
	{
		// Set the item to contain a bitmap
		pMenuInfo->hbmpUnchecked = CreateBitmapFromIcon(hWnd, iIconData, 14, 14);
	}

	// Do the insertation
	retValue = InsertMenuItem(hWinMenu,NULL,TRUE,pMenuInfo);


	// Return something to the upper levels
	return retValue;
}

HBITMAP CreateBitmapFromIcon(HWND hWnd, HICON hIcon, int width, int height)
{
	HDC     hCopyDC;
	HDC     hCompDC;
	HDC     hTempDC;
	HBITMAP hBitmap = NULL;
	HBITMAP hBitmapTemp = NULL;

	HGDIOBJ hGDIObj = NULL;
	HGDIOBJ hGDIObjTemp = NULL;

	// Get a DC for the window
	if (!(hCopyDC = GetDC(hWnd)))
	{
		// Failed to get the DC
	}
	else
	{
		hCompDC = CreateCompatibleDC( hCopyDC );
		hTempDC = CreateCompatibleDC( hCopyDC );

		// Create the two bitmaps for doing the paint and mask
		hBitmap = CreateCompatibleBitmap(hCopyDC, width, height);
		hBitmapTemp = CreateCompatibleBitmap(hCopyDC, width, height);

		// Select the bitmaps to draw on, do we need to replace
		// the GDI object since these are temps?!
		hGDIObj = SelectObject(hCompDC, hBitmap);
		hGDIObjTemp = SelectObject(hTempDC ,hBitmapTemp);

		// Draw the icon and the mask on seperate bitmaps, invert the mask
		// to remove the black areas
		DrawIconEx(hTempDC, -1, -1, hIcon, width, height, 0, NULL, DI_MASK);
		DrawIconEx(hCompDC , -1, -1, hIcon, width, height, 0, NULL, DI_IMAGE);

		// Combiend the mask and the image
		BitBlt(hCompDC, -1, -1, width, height, hTempDC, -1, -1, SRCINVERT);

		// Select the GDI objects so they get deleted
		SelectObject(hCompDC, hGDIObj);
		SelectObject(hTempDC, hGDIObjTemp);

		// Delete the temporary bitmaps
		DeleteObject(hBitmapTemp);

		// Delete the temporary device contexts
		DeleteDC(hCompDC);
		DeleteDC(hTempDC);
	}

	// Release the DC
	if (hCopyDC)
		ReleaseDC(hWnd, hCopyDC);

	// Return the bitmap or NULL
	return hBitmap;
}

BOOL DeleteMenuItems(HMENU hWinMenu)
{
	MENUITEMINFO iMenuItemInfo = { sizeof(MENUITEMINFO), MIIM_SUBMENU | MIIM_DATA | MIIM_CHECKMARKS , NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL };
	DWORD        iMenuCount = 0;
	DWORD        iCounter;

	// If the menu contains any items, look for memory we allocated
	if ((iMenuCount = GetMenuItemCount(hWinMenu)) > 0)
	{
		// Get the information for the menu item so we can free the 
		// memory that we allocated
		for (iCounter = 0; iCounter < iMenuCount; iCounter++)
		{
			if (!GetMenuItemInfo(hWinMenu,0,TRUE,&iMenuItemInfo))
			{
				// Didn't successfully get the menu item!?
			}
			else 
			{				
			    if (iMenuItemInfo.hSubMenu)
                {
                    // This is a sub menu, destroy him my robots
                    DeleteMenuItems(iMenuItemInfo.hSubMenu);
                    DestroyMenu(iMenuItemInfo.hSubMenu);
                }

				// Free the memory
				if (iMenuItemInfo.dwItemData)
                    free((void *)iMenuItemInfo.dwItemData);

				// Release the bitmap if there is one
				if (iMenuItemInfo.hbmpUnchecked)
					DeleteObject(iMenuItemInfo.hbmpUnchecked);
			}

			// Delete the menu item
			DeleteMenu(hWinMenu, 0, MF_BYPOSITION);
		}
	}

	return TRUE;
}

// Execute the menu item
BOOL ExecuteMenuItem(HMENU hWinMenu, DWORD iMenuItem)
{
	// The default menu item
	MENUITEMINFO iMenuItemInfo = { sizeof(MENUITEMINFO), MIIM_DATA, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL };
	HINSTANCE    hApp;
	LPSTR        parameterIndex = NULL;
	LPSTR 		 directoryIndex = NULL;

	// Get the menu item information
	if (GetMenuItemInfo(hWinMenu, iMenuItem, FALSE, &iMenuItemInfo))
	{
		if ((parameterIndex = strchr((const char *)iMenuItemInfo.dwItemData,',')) == NULL)
		{
			// No parameters
		}
		else 
		{
			// We have at least one parameter that we can pass to ShellExecute
			*parameterIndex = '\0';
			parameterIndex++;

			directoryIndex = strchr((const char *)parameterIndex,',');
		}

		// Check if the user was smart enough to pass in a startup directory as well
		if (directoryIndex != NULL)
		{
			// Append the \0 and move the directory pointer to the real string
			*directoryIndex = '\0';
			directoryIndex++;
		}

		// Call to execute using the parameters
		hApp = ShellExecute(NULL, "open", (const char *)iMenuItemInfo.dwItemData, parameterIndex, directoryIndex, SW_SHOWNORMAL);
	}

	// Return non-NULL (TRUE) if the application executed
	return (DWORD)hApp;
}

DllExport BOOL WBQuery (HWND hWnd)
{    
	return FALSE;
}

DllExport BOOL WBDraw (HWND hWnd, HDC DC, RECT* r, int state, BOOL active)
{
	// Bitmap our button will get copied into
	HDC       hCopyDC;
	HGDIOBJ   hGDIObj;

	// Make sure we have the menu button
	if (mMenuButton)
	{
		// Create an offscreen DC that we can work with
		hCopyDC = CreateCompatibleDC(DC);

		if (hCopyDC)
		{
			// Select our bitmap on the memory object
			hGDIObj = SelectObject(hCopyDC, mMenuButton);

			// Stretch blit our image on the new image
			StretchBlt(DC, r->left, r->top, (r->right - r->left), (r->bottom - r->top), hCopyDC, 0, 0, mMenuButtonData.bmWidth, mMenuButtonData.bmHeight, SRCCOPY);
			
			// Unselect our MenuButton object
			SelectObject(hCopyDC,hGDIObj);

			// Remove the temporary DC
			DeleteDC(hCopyDC);
		}
	}

    return TRUE;
}

